import logging
import time
import typing as t
import typer
import subprocess

from paraffin.utils import (
    get_custom_queue,
    get_stage_graph,
)
from paraffin.db import save_graph_to_db, get_job, complete_job

log = logging.getLogger(__name__)

app = typer.Typer()


@app.command()
def worker(
    queues: str = typer.Option(
        "default",
        "--queues",
        "-q",
        envvar="PARAFFIN_QUEUES",
        help="Comma separated list of queues to listen on.",
    ),
):
    """Start a Celery worker."""
    queues = queues.split(",")
    print(queues)
    while True:
        job = get_job(queues=queues)
        if job is None:
            # TODO: timeout
            print("No job found.")
            return
        try:
            subprocess.check_call(f"dvc repro -s {job['name']}", shell=True)
        except Exception as e:
            log.error(f"Failed to run job: {e}")
            complete_job(job["id"], status="failed")
        else:
            complete_job(job["id"], status="completed")


@app.command()
def submit(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    verbose: bool = typer.Option(False, help="Verbose output."),
):
    """Run DVC stages in parallel."""
    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    log.debug("Getting stage graph")
    graph = get_stage_graph(names=names, glob=True)
    print(f"Submitting {graph}")
    custom_queues = get_custom_queue()
    save_graph_to_db(graph, queues=custom_queues)



import logging
import os
import pathlib
import random
import shutil
import subprocess
import time

from celery import Celery

from paraffin.utils import clone_and_checkout, commit_and_push

log = logging.getLogger(__name__)


def make_celery() -> Celery:
    if url := os.environ.get("PARAFFIN_REDIS_URL"):
        app = Celery(
            __name__,
            broker=url,
            backend=url,
        )
    else:
        paraffin_folder = pathlib.Path(".paraffin")
        data_folder = paraffin_folder / "data"
        control_folder = paraffin_folder / "data"
        results_db = paraffin_folder / "results.db"
        gitignore = paraffin_folder / ".gitignore"

        data_folder.mkdir(parents=True, exist_ok=True)
        control_folder.mkdir(parents=True, exist_ok=True)
        results_db.parent.mkdir(parents=True, exist_ok=True)

        if not gitignore.exists():
            gitignore.write_text("data\nresults.db\n")

        app = Celery(
            __name__,
            broker_url="filesystem://",
            result_backend=f"db+sqlite:///{results_db.as_posix()}",
            broker_transport_options={
                "data_folder_in": data_folder.as_posix(),
                "data_folder_out": data_folder.as_posix(),
                "data_folder_processed": data_folder.as_posix(),
                "control_folder": control_folder.as_posix(),
            },
        )
    return app


app = make_celery()


def _run_dvc(self, name: str):
    """Run DVC repro command for a given stage.

    This task attempts to reproduce a specified DVC pipeline stage
    using the `dvc repro` command.
    If the command fails due to an "Unable to acquire lock" error,
    it retries the operation up to 5 times.
    If the error occurs after the stage has been executed, it attempts to
    commit the lock using the `dvc commit` command with a
    forced option to avoid loss of computational resources.
    """
    popen = subprocess.Popen(
        ["dvc", "repro", "--single-item", name],
        stdout=subprocess.PIPE,
        universal_newlines=True,
        stderr=subprocess.PIPE,
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        # logging.info(stdout_line)
        print(stdout_line, end="")
    popen.stdout.close()

    for stderr_line in iter(popen.stderr.readline, ""):
        # logging.error(stderr_line)
        print(stderr_line, end="")
        if "ERROR: Unable to acquire lock" in stderr_line:
            log.error(f"Retrying {name} due to lock error")
            raise self.retry(max_retries=5)
        if (
            f"ERROR: failed to reproduce '{name}': Unable to acquire lock"
            in stderr_line
        ):
            # unable to commit lock, keep retrying
            for _ in range(5):
                try:
                    log.error(f"Committing {name} again due to lock error")
                    subprocess.check_call(
                        ["dvc", "commit", name, "--force"],
                        stderr=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                    )
                    break
                except subprocess.CalledProcessError:
                    time.sleep(1 + random.random())
            else:
                raise RuntimeError(f"Unable to commit lock for {name}")
    popen.stderr.close()


def _run_vanilla(self, cmd: str):
    """Run a vanilla command for a given stage.

    This task attempts to run a specified command
    using the `subprocess.Popen` function.
    """
    print(f"Running command: {cmd}")
    subprocess.check_call(cmd, shell=True)


@app.task(bind=True, default_retry_delay=5)  # retry in 5 seconds
def repro(
    self,
    *args,
    name: str,
    branch: str,
    origin: str | None,
    commit: bool,
    cmd: str,
    use_dvc: bool,
):
    """Celery task to reproduce a DVC pipeline stage.

    Args:
        self (Task): The bound Celery task instance.
        *args: Additional arguments.
        name (str): The name of the DVC pipeline stage to reproduce.

    Raises:
        self.retry: If the "Unable to acquire lock" error occurs,
        the task is retried up to 5 times.
        RuntimeError: If unable to commit the lock after multiple attempts.

    Returns:
        bool: True if the operation is successful.
    """
    working_dir = pathlib.Path(os.environ.get("PARAFFIN_WORKING_DIRECTORY", "."))
    cleanup = True if os.environ.get("PARAFFIN_CLEANUP", "True") == "True" else False
    # print(f"Working directory: {working_dir} with cleanup: {cleanup}")

    if not working_dir.exists():
        working_dir.mkdir(parents=True)
    os.chdir(working_dir)

    clone_and_checkout(branch, origin)

    if use_dvc:
        _run_dvc(self, name)
    else:
        _run_vanilla(self, cmd)

    if commit:
        commit_and_push(name=name, origin=origin)

    if working_dir != pathlib.Path(".") and cleanup:
        # remove the working directory
        shutil.rmtree(working_dir)
    return True


@app.task(bind=True)
def skipped_repro(*args, **kwargs):
    """Dummy Celery task for testing purposes."""
    pass



from typing import Optional, List
from sqlmodel import Field, Relationship, SQLModel, create_engine, Session, select
import networkx as nx
from sqlalchemy.orm import joinedload
import fnmatch

class JobDependency(SQLModel, table=True):
    parent_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)
    child_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)

class Job(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    cmd: str  # Command to execute
    status: str = "pending"  # pending, running, completed, failed
    queue: str = "default"  # Queue name(s)

    # Relationships
    parents: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="children",
        sa_relationship_kwargs={"primaryjoin": "Job.id==JobDependency.child_id", "secondaryjoin": "Job.id==JobDependency.parent_id"}
    )
    children: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="parents",
        sa_relationship_kwargs={"primaryjoin": "Job.id==JobDependency.parent_id", "secondaryjoin": "Job.id==JobDependency.child_id"}
    )

def save_graph_to_db(graph: nx.DiGraph, queues: dict[str, str]):
    engine = create_engine("sqlite:///jobs.db")
    SQLModel.metadata.create_all(engine)
    with Session(engine) as session:
        for node in nx.topological_sort(graph):
            queue = "default"
            # use fnmatch to match the node name with the custom queues
            for pattern, q in queues.items():
                if fnmatch.fnmatch(node.name, pattern):
                    queue = q
                    break
            job = Job(cmd=node.cmd, name=node.name, queue=queue)
            session.add(job)
            # add dependencies
            for parent in graph.predecessors(node):
                parent_job = session.exec(select(Job).where(Job.cmd == parent.cmd)).one()
                session.add(JobDependency(parent_id=parent_job.id, child_id=job.id))
        
        session.commit()

def db_to_graph(db: str = "sqlite:///jobs.db") -> nx.DiGraph:
    engine = create_engine(db)
    with Session(engine) as session:
        jobs = session.exec(select(Job)).all()
        graph = nx.DiGraph()
        for job in jobs:
            graph.add_node(job.id, queue=job.queue, cmd=job.cmd, status=job.status)
            for parent in job.parents:
                graph.add_edge(parent.id, job.id)

    return graph


def get_job(db: str = "sqlite:///jobs.db", queues: list|None = None) -> dict | None:
    """
    Get the next job where status is 'pending' and all parents are 'completed'.
    """
    engine = create_engine(db)
    with Session(engine) as session:
        # Fetch jobs with 'pending' status, eagerly loading their parents
        statement = select(Job).where(Job.status == "pending")
        if queues:
            statement = statement.where(Job.queue.in_(queues))
        results = session.exec(statement)

        # Process each job to check if all parents are completed
        for job in results:
            print(len(job.parents))
            if all(parent.status == "completed" for parent in job.parents):
                job.status = "running"
                session.add(job)
                session.commit()
                return {
                    "id": job.id,
                    "name": job.name,
                    "cmd": job.cmd,
                    "queue": job.queue,
                    "status": job.status,
                }
    return None

def complete_job(job_id: int, db: str = "sqlite:///jobs.db", status: str = "completed"):
    engine = create_engine(db)
    with Session(engine) as session:
        statement = select(Job).where(Job.id == job_id)
        results = session.exec(statement)
        job = results.one()
        job.status = status
        session.add(job)
        session.commit()



import fnmatch
import pathlib
import subprocess

import dvc.api
import git
import networkx as nx
import yaml

from paraffin.abc import HirachicalStages


def get_subgraph_with_predecessors(graph, nodes, reverse=False):
    """
    Generate a subgraph containing the specified nodes and all their predecessors.

    Parameters
    ----------
    graph: networkx.DiGraph
        The original graph from which the subgraph is to be extracted.
    nodes: Iterable
        An iterable of nodes to be included in the subgraph along with
        their predecessors.
    reverse: bool, optional
        If True, the resulting subgraph will be reversed. Default is False.

    Returns
    -------
    networkx.Graph
        A subgraph containing the specified nodes and all their predecessors.
    """
    # Initialize a set to store nodes that will be in the subgraph
    nodes_to_include = set(nodes)

    # For each node in X, find all its predecessors
    for node in nodes:
        predecessors = nx.ancestors(graph, node)
        nodes_to_include.update(predecessors)

    # Create the subgraph with the selected nodes
    if reverse:
        return graph.subgraph(nodes_to_include).reverse(copy=True)
    return graph.subgraph(nodes_to_include).copy()


def get_stage_graph(names, glob=False) -> nx.DiGraph:
    """
    Generates a subgraph of stages from a DVC repository based on provided names.

    Attributes
    ----------
    names: list
        A list of stage names to filter the graph nodes.
    glob: bool, optional
        If True, uses glob pattern matching for names. Defaults to False.

    Returns
    -------
    networkx.DiGraph:
        A subgraph containing the specified stages and their predecessors.
    """
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    graph = fs.repo.index.graph.reverse(copy=True)
    nodes = [x for x in graph.nodes if hasattr(x, "name")]
    if names is not None and len(names) > 0:
        if glob:
            nodes = [
                x for x in nodes if any(fnmatch.fnmatch(x.name, name) for name in names)
            ]
        else:
            nodes = [x for x in nodes if x.name in names]

    subgraph = get_subgraph_with_predecessors(graph, nodes)

    # remove all nodes that do not have a name
    subgraph = nx.subgraph_view(subgraph, filter_node=lambda x: hasattr(x, "name"))

    return subgraph


def get_changed_stages(subgraph) -> list:
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    repo = fs.repo
    names = [x.name for x in subgraph.nodes]
    changed = list(repo.status(targets=names))
    graph = fs.repo.index.graph.reverse(copy=True)
    # find all downstream stages and add them to the changed list
    # Issue with changed stages is, if any upstream stage was changed
    # then we need to run ALL downstream stages, because
    # dvc status does not know / tell us because the immediate
    # upstream stage was unchanged at the point of checking.

    for name in changed:
        stage = next(x for x in graph.nodes if hasattr(x, "name") and x.name == name)
        for node in nx.descendants(graph, stage):
            changed.append(node.name)
    # TODO: split into definitely changed and maybe changed stages
    return changed


def get_custom_queue():
    try:
        with pathlib.Path("paraffin.yaml").open() as f:
            config = yaml.safe_load(f)

            return config.get("queue", {})

    except FileNotFoundError:
        return {}


def dag_to_levels(graph) -> HirachicalStages:
    """Converts a directed acyclic graph (DAG) into hierarchical levels.

    This function takes a directed acyclic graph (DAG) and organizes its nodes
    into hierarchical levels based on their distance from the root nodes.
    A root node is defined as a node with no predecessors.

    Arguments
    ---------
    graph: newtorkx.DiGraph
        A directed acyclic graph represented using NetworkX.

    Returns
    -------
    HirachicalStages
        A dictionary where the keys are levels (integers)
        and the values are lists of nodes at that level.

    Example:
        >>> import networkx as nx
        >>> G = nx.DiGraph()
        >>> G.add_edges_from([(1, 2), (1, 3), (3, 4)])
        >>> dag_to_levels(G)
        {0: [1], 1: [2, 3], 2: [4]}
    """
    nodes = []
    levels = {}
    for start_node in graph.nodes():
        if len(list(graph.predecessors(start_node))) == 0:
            if start_node not in nodes:
                for node in nx.bfs_tree(graph, start_node):
                    if node not in nodes:
                        nodes.append(node)
                        # find the longest path from the start_node to the current node
                        # to determine the level of the current node
                        level = 0
                        for path in nx.all_simple_paths(graph, start_node, node):
                            level = max(level, len(path) - 1)
                        try:
                            levels[level].append(node)
                        except KeyError:
                            levels[level] = [node]
                    else:
                        # this part has already been added
                        break
    return levels


def levels_to_mermaid(
    all_levels: list[HirachicalStages], changed_stages: list[str]
) -> str:
    # Initialize Mermaid syntax
    mermaid_syntax = "flowchart TD\n"

    for idx, levels in enumerate(all_levels):
        # Add each level as a subgraph
        for level, nodes in levels.items():
            mermaid_syntax += f"\tsubgraph Level{idx}:{level + 1}\n"
            for node in nodes:
                if node.name in changed_stages:
                    mermaid_syntax += f"\t\t{node.name}\n"
                else:
                    mermaid_syntax += f"\t\t{node.name}(✓)\n"
            mermaid_syntax += "\tend\n"

        # Add connections between levels
        for i in range(len(levels) - 1):
            mermaid_syntax += f"\tLevel{idx}:{i + 1} --> Level{idx}:{i + 2}\n"

    return mermaid_syntax


def clone_and_checkout(branch: str, origin: str | None) -> None:
    # check if we are in a git repo
    try:
        repo = git.Repo()
        if origin is not None:
            if origin != str(repo.remotes.origin.url):
                raise ValueError(
                    f"Origin mismatch: {origin} != {str(repo.remotes.origin.url)}"
                )
        if branch != str(repo.active_branch):
            repo.git.checkout(branch)
    except git.InvalidGitRepositoryError:
        if origin is None:
            raise ValueError("Cannot clone a repository without an origin.")
        print(f"Cloning {origin} into current directory.")
        repo = git.Repo.clone_from(origin, ".")
        print(f"Checking out branch {branch}.")
        repo.git.checkout(branch)
    if origin is not None:
        print("Pulling latest changes.")
        repo.git.pull("origin", branch)
        subprocess.check_call(["dvc", "pull"])


def commit_and_push(name: str, origin) -> None:
    repo = git.Repo()
    if repo.is_dirty():
        print("Committing changes.")
        repo.git.add(".")
        repo.git.commit("-m", f"paraffin: auto-commit {name}")
        if origin is not None:
            print("Pushing changes.")
            repo.git.push("origin", repo.active_branch)
            subprocess.check_call(["dvc", "push"])
